#!/usr/bin/perl -w

##
# This script checks the links of the website given on the command line.
#
# See printUsage() message for more details.
#
# Created: Zach Tomaszewski, 15 Apr - 06 May 2004
##

use strict;         #included modules
use Getopt::Std;
use IO::Socket;
use URI;

my $EOL = "\015\012";  #constants
my $rootURL = "";      #global variables
my %checkedURLs;   #those URLs we've retrieved; values contain response codes

use vars qw/ $opt_V $opt_H $opt_r $opt_d $opt_q $opt_b $opt_f/;
Getopt::Std::getopts('VHdqr:b:f:');
if ($opt_H && !$opt_V) {
  $opt_V = 1;
}

main();             #call main to start program


##
# Checks command line args,
# gets root url from command line,
# and starts processing.
#
# Returns 0 if processing successful; 1 on an error.
#
sub main {
  if (@ARGV != 1) {
    printUsage();
    return 1;
  }else {
    #get valid root URL
    my $firstURL = URI->new($ARGV[0], 'http'); 
    if (!validateURL($firstURL->as_string, 'http', 'https')) {
      print "ERROR: Invalid HTTP URL given ($firstURL)\n";
      return;
    }    
    
    if (defined $opt_r) { #should recurse even on a file, so root is the directory
      #chop off any end file from firstURL (use greedy operator and backtrack)
      $rootURL = ($firstURL =~ m!(https?://.*/)[^/]+$!i) ? $1 : $firstURL;
    }else {
      $rootURL = $firstURL;
    }
    print("[checkweb] Root URL: $rootURL\n") if ($opt_V);

    #start processing first URL
    checkLinks($firstURL, 0);
  }
}


##
# Prints a usage message to the screen.
#
sub printUsage {
  print qq^ 
  Usage: checkweb [-VHdq] [-r #] [-b #] [-f 'search phrase'] rooturl 
  
  Checkweb will check for broken links in a web page.
  
  Checkweb will retreive the given rooturl and attempt to parse it as an
  HTML page.  It will then check that each HTML <a href> or <img src> link 
  from that page returns a successful HTTP response.  If rooturl is a 
  directory (or the -r switch is used), checkweb will recursively parse HTML 
  pages found through links to the rooturl's directory or subdirectories 
  (that is, it will check the links for the whole site rather than just the 
  page).  
  
  Example rooturls:
    http://domain.com/img.jpg -- check that file can be retrieved
    http://domain.com/index.html -- check all links found in this page
    http://domain.com/ -- check links in page found at domain.com/ and also
                          check links of all linked domain.com site pages.

  Options:
    -b#  BREADTH.  Make only this many original requests, then stop. 
    -d   DON'T recurse, even when given a DIRECTORY. (Overrides -r#)  
    -f   FIND and count the given search phrase in each HTML page parsed.
    -H   Print HTTP response HEADERS.  (Also turns on -V.)
    -q   QUIET -- don't print successfully retrieved urls.
    -r#  Treat the given url as the root page of a website, and RECURSIVELY
         check all site HTML pages linked to from the rooturl file.
         Will recurse only to a depth of #, or unlimited if # is 0.
    -V   VERBOSE (aka, debug info)
  \n^;
}


##
# validateUrl ( $url , [@proto] )
# 
# Checks that the given $url string matches the form:
#   protocol : host [ :port ] [/path]
#
# If @proto is given, will verify that the $url's protocol 
# also matches one of the passed permissible protocols.
#
# Returns true (>0) if valid, false (0) otherwise
#
sub validateURL {
  my($url, @proto) = @_;
  my $valid = 0;
  if ($url =~ m!^[\w.-]+:(//)?   #proto
                \w+(:\d+)?       #host and opt. port#
                /?([\w./-]+)*  #path                 
               !ix){
               
    #valid url so far, so check protos if given
    if (@proto) {
      foreach my $p (@proto) {
        if ($url =~ m!^$p!) { #starts with one of given protos
          $valid = 1;
        }
      }
    }            
  }
  return $valid;
}


##
# checkLinks ( $url, [$indent_level], [$is_a_redirected_url] )
#
# Will download the given $url.  
# If it is an HTML page, it will parse it for more links to 
#  check with checkLinks.  Will not recurse on links not hierarchically
#  beneath $rootURL
# If not an HTML doc, it will just verify that the resource can be successfully
#  retrieved.
#
# $indent_level is used for printing to determine how deeply recursed we are.
#   It should start at 0 (or be omitted) on the first call, and be ++ each
#   time checkLinks is called.
# If $is_a_redirected_url, will report the url as such ("Redirected to: ...")
#
sub checkLinks {
  my ($url, $indent, $redirected) = @_;
  my ($get_content) = 0;
  
  #first see if we even need to check this URL; return early if not
  if ($opt_b && keys %checkedURLs >= $opt_b){
    return;  #checked breadth urls already
  }
  if (defined $checkedURLs{$url}) { #checked this url before
    printResponse($url, $indent, $checkedURLs{$url}, 
                  "[Code from original request.]");
    return;
  }elsif (!validateURL($url->as_string, 'http', 'https') ) { 
    printResponse($url, $indent, 0, "Skipped non-HTTP URL.");
    return;
  }    

  #open and check socket
  my $sock = IO::Socket::INET->new(PeerAddr => $url->authority,
                             PeerPort => $url->port,
                             Proto    => 'tcp');
  unless ($sock) {
    printResponse($url, $indent, 0, "Connection Failed.");
    return;
  }

  if ($url->as_string =~ /^$rootURL/) { #url is under root, need to GET content
    $get_content = 1;
  }

  #verbose/debug info
  if ($opt_V) { 
    print "[checkweb] ";
    print (($get_content) ? "GET " : "HEAD ");
    print (($url->path) ? $url->path : "/");
    print " from " . $url->authority . "\n";
  }
  
  #request page
  print $sock (($get_content) ? "GET " : "HEAD ");
  print $sock (($url->path) ? $url->path : "/");
  print $sock (" HTTP/1.1" . $EOL);
  print $sock ("Host: " . $url->authority . $EOL);
  print $sock ("Connection: close" . $EOL);
  print $sock ($EOL);
  $sock->flush();
  
  #receive and process response
  my @content = "";
  my $inheader = 1;  #start off processing header
  my $length = 0;    #reported Content-Length
  my $bytecount = 0; #message body length
  my $content_type = 0;
  my $redirect_to = 0;
  while ( <$sock> ) {  
    if ($inheader) {
      #look for certain lines and handle them
      if (/^HTTP\/\d\.\d (\d\d\d) (.*)$EOL/){ #get and handle response code
        $checkedURLs{$url} = $1;  #save response code
        if (!$redirected) {   #normal case
          printResponse($url, $indent, $1, $2);
        }else {
          #this url is a redirected url, so differen output
          printResponse("Redirected to: $url", $indent, $1, $2, 600);
        }
      }

      if (/^Content-Type: *(.+)$EOL/i){         #record content type
        $content_type = $1;      
      }elsif(/^Content-Length: *(\d+)$EOL/i){   #record length
        $length = $1;
      }elsif(/^Location: *(.+)$EOL/i && int($checkedURLs{$url}/100) == 3){
        $redirect_to = $1;
      }elsif (/^$EOL$/) {                       #blank line -> end of header
        $inheader = 0;
        if (!$get_content) { #have all we need, since only did HEAD
          close($sock);
          last;
        }
      }else {
        if ($opt_H) {print " $_";}  #print headers for debugging
      }        
    }else {
      #receiving message body; increase
      @content = (@content, $_);
      $bytecount += length($_);
      if ($length && $bytecount == $length) {
        close($sock); #got everything, so close rather than wait for time out
        last;
      }
    }
  }
  close $sock;

  #report on any found strings
  if ($opt_f){
    my $string_count = 0;
    my $content = join("", @content); #to support searches across lines
    while ($content =~ /$opt_f/oig){
      $string_count++;
    }
    if ($string_count) {
      #XXX: tricky, hacky use of printResponse
      printResponse("Search string found", $indent, $string_count, "times.", 600);
    }
  }

  #now process the results for successful HTML pages (only)
  if ($content_type && $content_type =~ m!text/html!
      && int($checkedURLs{$url}/100) == 2  && !$opt_d) {
    if ($url->as_string =~ /^$rootURL/ &&  #only if beneath root
       ($opt_r) ? $opt_r > $indent : 1) { #and any recurse limit not yet met
      parseForLinks($url, $indent, @content);
    }
  }
  
  #if we got a redirect, then go a head and check the redirected URL
  #using standard checkLinks behavior
  if ($redirect_to){
    $redirect_to = URI->new($redirect_to);
    checkLinks($redirect_to, $indent, 1);  
  }
}


##
# printResponse ( $url, $indent, $response_code, [$response_message], 
#                  [$indent_style])
#
# Prints <indenting> <URL> -> <response> <message>
# Substitutes correct message for given code if no message given.
# Does not print response codes of 0.
# A response code given in $indent_style will override $response_code
#  for purposes of indenting style.
# Will print not print successful responses (200) if $opt_q is set.
#
sub printResponse {
  my ($url, $indent, $response_code, $response_message, $indent_style) = @_;

  #return code at a glance: no response, 100, 200, ..., 500, extra line info
  my @indentstr = ("?? ", "^^ ", "-- ", "== ", "++ ", "!! ", "   ");

  if ($opt_q && $response_code == 200) {
    return;
  }
  if (!defined $indent_style) {
    $indent_style = $response_code;
  }

  use integer; #integer math until end of block
  print ($indentstr[$indent_style/100] x $indent);
  print ($url . " -> ");
  if ($response_code) { 
    print $response_code . " "; 
  }
  if (defined $response_message) {  #use the one given
    print $response_message;
  }else {  #print corresponding GENERAL error code
    #pretend switch
    if ($response_code/100 == 200/100) { print "OK.";}
    elsif ($response_code/100 == 100/100) { print "Informational response.";}
    elsif ($response_code/100 == 300/100) { print "Redirected.";}
    elsif ($response_code/100 == 400/100) { print "Inaccessible/Not Found.";}
    elsif ($response_code/100 == 500/100) { print "Server Error.";}
    else { print "Unknown error."; }
  }
  print "\n";
}


##
# parseForLinks ($current_url, $indent, @message_body_lines)
#
# Searches the lines given for HTML anchors <a> and <img> tags.
# (XXX: could also search for LINKs, OBJECTs, and APPLETs someday)
#
sub parseForLinks {
  my ($currentURL, $indent, @file_lines) = @_;
  my ($current_dir) = ($currentURL =~ m!(http://.*/)[^/]+$!i) ? $1 : $currentURL;
  my ($all_lines) = join("", @file_lines);  #for searches across lines
  
  #find all <a hrefs>
  my (@found_urls) = ($all_lines =~ /<\s*a[^>]+  #opening of tag until href part
                                      href\s*=\s*  
                                      "?\s*        # " and white space optional
                                      ([^>"]+)     #acutal URL to save
                                      \s*"?[^>]*>    #make sure we don't leave tag
                                    /xig);
  #add all <img srcs>                                
  push (@found_urls, ($all_lines =~ /<\s*img[^>]+  #opening of tag until src part
                                      src\s*=\s*  
                                      "?\s*        # " and white space optional
                                      ([^>"]+)     #acutal URL to save
                                      \s*"?[^>]*>    #make sure we don't leave tag
                                    /xig) );  
  #reduce to unique urls only
  my($url, %unique_urls);
  foreach $url (@found_urls){
    #remove fragments
    $url =~ s/#.*//;  #will leave fragments as ""
    if ($url eq "") {
      next;
    }
    $unique_urls{$url} = 1;
  }
  @found_urls = keys %unique_urls;  

  #debug/verbose printing of urls before processing starts
  if ($opt_V) {
    foreach $url (@found_urls){
      print "[checkweb] Raw url found: $url \n";
    }
  }
  
  #process each found link
  foreach $url (@found_urls){
    $url = new URI($url); #turn url into an object
    $url = $url->abs($current_dir);  #create an absolute URL if not already
    checkLinks($url, $indent + 1);
  }  

}# end parseForLinks sub

